--------------------------------------------------
-- Mir AI v1.0 (Beta 2c) by Miranda Blade
--------------------------------------------------

require "./AI/USER_AI/Const.lua"
require "./AI/USER_AI/Util.lua"
require "./AI/USER_AI/Config.lua" -- configuration file
require "./AI/USER_AI/Patrol.lua"

--------------------------------------------------
-- ############ INITIALIZATION ############
--------------------------------------------------

--------------------------------------------------
-- State (from default AI)
--------------------------------------------------
IDLE_ST              = 0
FOLLOW_ST            = 1
CHASE_ST             = 2
ATTACK_ST            = 3
MOVE_CMD_ST          = 4
STOP_CMD_ST          = 5
ATTACK_OBJECT_CMD_ST = 6
ATTACK_AREA_CMD_ST   = 7
PATROL_CMD_ST        = 8
HOLD_CMD_ST          = 9
SKILL_OBJECT_CMD_ST  = 10
SKILL_AREA_CMD_ST    = 11
FOLLOW_CMD_ST        = 12
--------------------------------------------------
-- State (extra)
--------------------------------------------------
EVADE_ST             = 50
   EVADE_COMEBACK       = 2
--------------------------------------------------
-- "Output movements" (they notify the user about some events)
--------------------------------------------------
OUTMOVE_NEWFRIEND_ST = 60 -- The choosen player has been added to the friend-list
OUTMOVE_DELFRIEND_ST = 61 -- " has been removed

--------------------------------------------------
-- Skill delays (milliseconds)
--------------------------------------------------
DELAY_AAA      = 500 -- Cart Revolution and Bash (Cultus)
DELAY_AAA_BOLT = 3000 -- extra delay for weapon bolts
DELAY_AAP      = 1000 -- Auto Aid Potion

DELAY_AMI_BULW = 1000 -- Bulwark lasts 45 - (skill lvl *5) seconds (lvl 5 = 20")
DELAY_FIL_ACCL = 120000 -- Accel. Flight lvl 5 lasts 120"
DELAY_LIF_HEAL = 1000
DELAY_LIF_ESCP = 40000 -- Urgent Escape lvl 5 lasts 40"
DELAY_VAN_BLES = 5000

DELAY_SLOW_POWER = 2000 -- for slow_power skill mode

--------------------------------------------------
-- Skill activation and time-outs (global var.)
--------------------------------------------------
AAP_Engaged = false -- it becomes true when an auto aip potion is casted
AAP_TimeOut = 0     -- time (tick) when AI can cast another AAP

AAA_Engaged = false -- Alchemist auto-attack: can be cart revolution (multiple target) or weapon based skill (single targets)
AAA_TimeOut = 0

AS_AMI_BULW_Engaged = false -- is Amistr Bulwark still on?
AS_AMI_BULW_TimeOut = 0 -- time (tick) when the skill goes off an can be casted again

AS_FIL_ACCL_Engaged = false
AS_FIL_ACCL_TimeOut = 0

AS_LIF_ESCP_Engaged = false
AS_LIF_ESCP_TimeOut = 0

AS_LIF_HEAL_Engaged = false
AS_LIF_HEAL_TimeOut = 0

AS_VAN_BLES_Engaged = false
AS_VAN_BLES_TimeOut = 0

--------------------------------------------------
-- Global variable (dafaul AI)
--------------------------------------------------
MyState      = IDLE_ST
MyEnemy      = 0
MyDestX      = 0
MyDestY      = 0
MyPatrolX    = 0
MyPatrolY    = 0
ResCmdList   = List.new()
MyID         = 0
MySkill      = 0
MySkillLevel = 0
--------------------------------------------------
-- Global variable (extra)
--------------------------------------------------
CircleDir            = 1 -- Current circle-move direction
CircleBlockTime      = 0 -- Time when position started to not change (obstacle encountered?)
CircleBlockStep      = 0 -- 0 = all ok; 1 = obstacle? Check a second; 2 = obstacle detected
vHomunX              = 0 --\__ I'm really moving?
vHomunY              = 0 --/
vHomunHP             = 0
AtkStartTime         = 0
AtkSkillDoneCount    = 0
Evade_ComeBackTime   = 0
BlockTime            = 0
BlockFound           = 0
BlockedPathToEnemyID = 0
OutMoveTime          = 0
OutMoveTarget        = 0
FriendList_Loaded    = false

TactBeha = -1
TactSkill = -1

--------------------------------------------------
-- ############ COMMAND PROCESS ###########
--------------------------------------------------

--------------------------------------------------
function	OnMOVE_CMD(x, y)
--------------------------------------------------
	TraceAI("OnMOVE_CMD")

	if (x == MyDestX and y == MyDestY and MOTION_MOVE == GetV(V_MOTION, MyID)) then
		return
	end

   -- are you adding/removing a friend? ----------
	local actors = GetActors()
	local OwnerID = GetV(V_OWNER, MyID)
   for i,v in ipairs(actors) do
      if (v ~= OwnerID and v ~= MyID) then
         if (0 == IsMonster(v)) then
            local PlayerX, PlayerY = GetV(V_POSITION, v)
            if (PlayerX == x) and (math.abs(PlayerY-y) <=1) then
               FriendList_Switch(v)             -- update friend list
            	OutMoveTime = GetTick()          -- notify user
            	OutMoveTarget = v
               if isNotFriend(v) then
               	MyState = OUTMOVE_DELFRIEND_ST
               else
                  MyState = OUTMOVE_NEWFRIEND_ST
               end
            	MyDestX = x
            	MyDestY = y
            	MyEnemy = 0
               MySkill = 0
               return
            end
         end
      end
   end

	local curX, curY = GetV(V_POSITION, MyID)
	if (math.abs(x-curX) + math.abs(y-curY) > 15) then
		List.pushleft (ResCmdList,{MOVE_CMD,x,y})
		x = math.floor((x+curX)/2)
		y = math.floor((y+curY)/2)
	end

	Move(MyID, x, y)

	MyState = MOVE_CMD_ST
	MyDestX = x
	MyDestY = y
	MyEnemy = 0
	MySkill = 0

end

--------------------------------------------------
function	OnSTOP_CMD()
--------------------------------------------------
	TraceAI("OnSTOP_CMD")

	if (GetV(V_MOTION,MyID) ~= MOTION_STAND) then
		Move(MyID, GetV(V_POSITION,MyID))
	end
	MyState = IDLE_ST
	MyDestX = 0
	MyDestY = 0
	MyEnemy = 0
	MySkill = 0

end

--------------------------------------------------
function	OnATTACK_OBJECT_CMD(id)
--------------------------------------------------
	TraceAI("OnATTACK_OBJECT_CMD")

	MySkill = 0
	MyEnemy = id
	MyState = CHASE_ST

end

--------------------------------------------------
function	OnATTACK_AREA_CMD(x, y)
--------------------------------------------------
	TraceAI("OnATTACK_AREA_CMD")

	if (x ~= MyDestX or y ~= MyDestY or MOTION_MOVE ~= GetV(V_MOTION, MyID)) then
		Move(MyID, x, y)
	end
	MyDestX = x
	MyDestY = y
	MyEnemy = 0
	MyState = ATTACK_AREA_CMD_ST

end

--------------------------------------------------
function	OnPATROL_CMD(x, y)
--------------------------------------------------
	TraceAI ("OnPATROL_CMD")

	MyPatrolX , MyPatrolY = GetV(V_POSITION, MyID)
	MyDestX = x
	MyDestY = y
	Move (MyID,x,y)
	MyState = PATROL_CMD_ST

end

--------------------------------------------------
function	OnHOLD_CMD()
--------------------------------------------------
	TraceAI("OnHOLD_CMD")

	MyDestX = 0
	MyDestY = 0
	MyEnemy = 0
	MyState = HOLD_CMD_ST

end

--------------------------------------------------
function	OnSKILL_OBJECT_CMD(level, skill, id)
--------------------------------------------------
	TraceAI("OnSKILL_OBJECT_CMD")

	MySkillLevel = level
	MySkill = skill
	MyEnemy = id
	MyState = CHASE_ST

end

--------------------------------------------------
function	OnSKILL_AREA_CMD(level, skill, x, y)
--------------------------------------------------
	TraceAI("OnSKILL_AREA_CMD")

	Move(MyID,x,y)
	MyDestX = x
	MyDestY = y
	MySkillLevel = level
	MySkill = skill
	MyState = SKILL_AREA_CMD_ST

end

--------------------------------------------------
function	OnFOLLOW_CMD()
--------------------------------------------------
	if (MyState ~= FOLLOW_CMD_ST) then
		MoveToOwner(MyID)
		MyState = FOLLOW_CMD_ST
		MyDestX, MyDestY = GetV(V_POSITION, GetV(V_OWNER, MyID))
		MyEnemy = 0
		MySkill = 0
		TraceAI("OnFOLLOW_CMD")
	else
		MyState = IDLE_ST
		MyEnemy = 0
		MySkill = 0
		TraceAI("FOLLOW_CMD_ST --> IDLE_ST")
	end

end

--------------------------------------------------
function	ProcessCommand(msg)
--------------------------------------------------
	if     (msg[1] == MOVE_CMD) then
		OnMOVE_CMD(msg[2], msg[3])
		TraceAI("MOVE_CMD")
	elseif (msg[1] == STOP_CMD) then
		OnSTOP_CMD()
		TraceAI("STOP_CMD")
	elseif (msg[1] == ATTACK_OBJECT_CMD) then
		OnATTACK_OBJECT_CMD(msg[2])
		TraceAI("ATTACK_OBJECT_CMD")
	elseif (msg[1] == ATTACK_AREA_CMD) then
		OnATTACK_AREA_CMD(msg[2], msg[3])
		TraceAI("ATTACK_AREA_CMD")
	elseif (msg[1] == PATROL_CMD) then
		OnPATROL_CMD(msg[2], msg[3])
		TraceAI("PATROL_CMD")
	elseif (msg[1] == HOLD_CMD) then
		OnHOLD_CMD()
		TraceAI("HOLD_CMD")
	elseif (msg[1] == SKILL_OBJECT_CMD) then
		OnSKILL_OBJECT_CMD(msg[2], msg[3], msg[4], msg[5])
		TraceAI("SKILL_OBJECT_CMD")
	elseif (msg[1] == SKILL_AREA_CMD) then
		OnSKILL_AREA_CMD(msg[2], msg[3], msg[4], msg[5])
		TraceAI ("SKILL_AREA_CMD")
	elseif (msg[1] == FOLLOW_CMD) then
		OnFOLLOW_CMD()
		TraceAI("FOLLOW_CMD")
	end
end

--------------------------------------------------
-- ############# STATE PROCESS ############
--------------------------------------------------

--------------------------------------------------
function	DoSkill_AidPotion(AAPTargetID)
--------------------------------------------------
   if (AAP_Mode > 0) then
      if (AAP_Engaged) then                     -- if AAP was just casted, wait for a little
         if (GetTick() > AAP_TimeOut) then
            AAP_Engaged = false
         end
      else                                      -- else it is the time to cast it
         MySkill = 231
         MySkillLevel = AAP_Level
         AAP_TimeOut = GetTick() + DELAY_AAP    -- 1" delay
         AAP_Engaged = true
         SkillObject(MyID, MySkillLevel, MySkill, AAPTargetID)
         MySkill = 0
      end
   end
end

--------------------------------------------------
function	DoSkill_AutoAttack(SkillID)
--------------------------------------------------
   local OwnerID = GetV(V_OWNER, MyID)
   local OwnerSP = GetV(V_SP, OwnerID)
   local OwnerHP = GetV(V_HP, OwnerID)
   if (AAA_Engaged) then                        -- if an AAA skill was just casted, wait for a little
      if (GetTick() > AAA_TimeOut) then
         AAA_Engaged = false
      end
   else                                         -- else it is the time to cast it
      if (OwnerHP > AAA_MinHP and OwnerHP < AAA_MaxHP) then
         if(((SkillID ~= 153) and (OwnerSP >= AST_MinSP) )
         or ((SkillID == 153) and (OwnerSP >= ACR_MinSP) ))
         then                  -- preserve some SP
            local ExtraDelay = 0
            if ((SkillID ~= 153) and (SkillID ~= 5)) then -- it is not Bash nor Cart R.
               ExtraDelay = ExtraDelay + DELAY_AAA_BOLT
            end
            MySkillLevel = 10
            AAA_TimeOut = GetTick() + DELAY_AAA + ExtraDelay
            AAA_Engaged = true
            local object = GetOwnerEnemy(MyID)
            if (object > 0) then
               MyState = CHASE_ST
               MyEnemy = object
            end
            local EnemyType = GetV(V_HOMUNTYPE, MyEnemy)
            if (EnemyType < 1078 or EnemyType > 1085) then -- don't waste SP on plants and mushrooms
               MySkill = SkillID
               TraceAI(string.format("Doing autoattack skill on enemy type %d",EnemyType))
               SkillObject(MyID, MySkillLevel, MySkill, MyEnemy)
               MySkill = 0
            end
         end
      end
   end
end

--------------------------------------------------
function	DoSkill_HealingTouch(HealTargetID)
--------------------------------------------------
   local HomunSP = GetV(V_SP,MyID)
   if (AS_LIF_HEAL_Engaged) then                -- if HT was just casted, wait a little
      if (GetTick() > AS_LIF_HEAL_TimeOut) then
         AS_LIF_HEAL_Engaged = false
      end
   else                                         -- else it is the time to cast it
      if (HomunSP >= AUTOSKILL_SP_LIF_HEAL) then -- if there are enough SP left
         MySkill = 8001
         MySkillLevel = 5
         AS_LIF_HEAL_TimeOut = GetTick() + DELAY_LIF_HEAL
         AS_LIF_HEAL_Engaged = true
         SkillObject(MyID, MySkillLevel, MySkill, HealTargetID)
      end
   end
end

--------------------------------------------------
function	DoSkill_UrgentEscape()
--------------------------------------------------
   local HomunSP = GetV(V_SP,MyID)
   if (AS_LIF_ESCP_Engaged) then             -- if UE is already active, wait until it goes OFF
      if (GetTick() > AS_LIF_ESCP_TimeOut) then
         AS_LIF_ESCP_Engaged = false
      end
   else                                      -- if UE is OFF, activate it
      if (HomunSP >= AUTOSKILL_SP_LIF_ESCP) then -- if there are enough SP left
         MySkill = 8002
         MySkillLevel = 5
         AS_LIF_ESCP_TimeOut = GetTick() + DELAY_LIF_ESCP
         AS_LIF_ESCP_Engaged = true
         SkillObject(MyID, MySkillLevel, MySkill, MyID)
      end
   end
end

--------------------------------------------------
function	DoSkill_CBlessing()
--------------------------------------------------
   local HomunSP = GetV(V_SP,MyID)
   if (AS_VAN_BLES_Engaged) then             -- if CB is already active, wait until it goes OFF
      if (GetTick() > AS_VAN_BLES_TimeOut) then
         AS_VAN_BLES_Engaged = false
      end
   else                                      -- if CB is OFF, activate it
      if (HomunSP >= AUTOSKILL_SP_VAN_BLES) then -- if there are enough SP left
         MySkill = 8014
         MySkillLevel = 4 -- heals chance: 36% enemy, 60% self, 4% owner
         AS_VAN_BLES_TimeOut = GetTick() + DELAY_VAN_BLES
         AS_VAN_BLES_Engaged = true
         SkillObject(MyID, MySkillLevel, MySkill, MyID)
      end
   end
end

--------------------------------------------------
function	CircleAroundTarget(TrgID)
--------------------------------------------------

   local OwnerID        = GetV(V_OWNER, MyID)
   local OwnerX, OwnerY = GetV(V_POSITION, OwnerID)
   local TrgX, TrgY     = GetV(V_POSITION, TrgID)
   local x, y           = GetV(V_POSITION, MyID)

   --------------- current destination -----------
   if (Evade_ComeBackTime == EVADE_COMEBACK) or (CircleBlockStep == 2) then
      MyDestX = OwnerX
      MyDestY = OwnerY
   else
      MyDestX = TrgX + AAI_CIRC_X[CircleDir]       -- direction defined by the circle step
      MyDestY = TrgY + AAI_CIRC_Y[CircleDir]
   end

   --------------- obstacle detection ------------
   if (math.abs(vHomunX - x) == 0)
   and(math.abs(vHomunY - y) == 0) then         -- if there is no movement
      if (CircleBlockStep == 0) then
         CircleBlockTime = GetTick() + 500      -- 500 ms time-out
         CircleBlockStep = 1
      elseif (CircleBlockStep == 1) then        -- it is in wait mode until the time out expires
         if (GetTick () > CircleBlockTime) then -- if it expires
            CircleBlockStep = 2                 -- we assume that an obstacle was found in the current direction
         end
      end
   else
      CircleBlockStep = 0
   end

   --------------- position reached? -------------
   if  (math.abs(MyDestX - x) < 1)
   and (math.abs(MyDestY - y) < 1) -- yes!
   then                            -- select next direction
      if (MyState == EVADE_ST) then
         if CircleBlockStep == 2 then
            CircleDir = CircleDir + 1
            Evade_ComeBackTime = 0
         else
            if (Evade_ComeBackTime == EVADE_COMEBACK) then
               if (GetDistance2(MyEnemy, OwnerID) <=1) then
                  CircleDir = CircleDir + 2        -- the mosnter shoud reach the owner
                  Evade_ComeBackTime = 0
               end
            else
               CircleDir = CircleDir + 2
               Evade_ComeBackTime = Evade_ComeBackTime + 1
            end
         end
      else
         CircleDir = CircleDir + 1
         Evade_ComeBackTime = 0
      end
      if (CircleDir > AAI_CIRC_MAXSTEP) then
         CircleDir = 1
      end
   end

   ----- record curren position and move on ------
   vHomunX = x
   vHomunY = y
	Move(MyID, MyDestX, MyDestY)
	return
end

--------------------------------------------------
function	OnIDLE_ST()
--------------------------------------------------
	-- TraceAI("OnIDLE_ST")

	local cmd = List.popleft(ResCmdList)
	if (cmd ~= nil) then
		ProcessCommand(cmd)
		return
	end

	AtkStartTime = 0
	AtkSkillDoneCount = 0
   TactBeha = -1
   TactSkill = -1

   local HomunHP = GetV(V_HP, MyID)
	local HomunMaxHP = GetV(V_MAXHP, MyID)
	local HomunHPPerc = (HomunHP / HomunMaxHP) * 100
   local HomunSP = GetV(V_SP, MyID)
	local HomunMaxSP = GetV(V_MAXSP, MyID)
   local type = GetV(V_HOMUNTYPE, MyID)

   if (HomunHP < HomunMaxHP) then
      if(type == VANILMIRTH  or type == VANILMIRTH_H
      or type == VANILMIRTH2 or type == VANILMIRTH_H2) then
         DoSkill_CBlessing()
      end
   end

   -- [switch to attack] conditions: -------------
	local object = nil

   -- 1: if the homunculus is under attack
   --> but evade if HP < secure level
   --> else engage the enemy
	object = GetMyEnemy_AttackingMe(MyID) -- get TactBeha and TactSkill too
	if (object ~= 0) then
   	if ( (HomunHPPerc < HP_PERC_DANGER) or (TactBeha == BEHA_avoid) or (TactBeha == BEHA_coward) ) then
		   MyState = EVADE_ST
         MyEnemy = object
         Evade_ComeBackTime = EVADE_COMEBACK
         TraceAI("IDLE_ST -> EVADE_ST: attacked while HP < HP_PERC_DANGER")
         return
      else
         MyState = CHASE_ST
         MyEnemy = object
         TraceAI("IDLE_ST -> CHASE_ST: attacked, intercepting agressor")
         return
      end
   end

   -- 2: if the owner is under attack --> help owner
	object = GetOwnerEnemy(MyID)
   if (object ~= 0) then
	   TactBeha = GetTact(V_TACTBEHAV, GetV(V_HOMUNTYPE, object))
	   TactSkill= GetTact(V_TACTSKILL, GetV(V_HOMUNTYPE, object))
      if (TactBeha ~= BEHA_avoid) then
         MyState = CHASE_ST
         MyEnemy = object
         TraceAI("IDLE_ST -> CHASE_ST: owner under attack")
         return
      end
   end

   -- 3: if the owner or a friend is attacking --> cooperate
   local CooTarget = 0
   object = GetV(V_TARGET,GetV(V_OWNER, MyID))
   OwnerMotion = GetV(V_MOTION, GetV(V_OWNER, MyID))
	if (object ~= 0) and (OwnerMotion == MOTION_ATTACK or OwnerMotion == MOTION_ATTACK2) then
	  CooTarget = object
	end
	if CooTarget == 0 then -- Owner is not attackin. Maybe a friend?
      local FriendMotion = 0
      for i,v in Friends do
         object = GetV(V_TARGET, v)
         FriendMotion = GetV(V_MOTION, v)
      	if (object ~= 0) and (FriendMotion == MOTION_ATTACK or FriendMotion == MOTION_ATTACK2) then
            CooTarget = object
            break
         end
      end
	end
   if CooTarget ~= 0 then
	   TactBeha = GetTact(V_TACTBEHAV, GetV(V_HOMUNTYPE, CooTarget))
	   TactSkill= GetTact(V_TACTSKILL, GetV(V_HOMUNTYPE, CooTarget))
      if (TactBeha ~= BEHA_avoid) then
         MyState = CHASE_ST
         MyEnemy = CooTarget
         TraceAI("IDLE_ST -> CHASE_ST: cooperative attack")
         return
      end
	end

   -- 4: aggressive, if HPs are OK and a target is near
   if (HomunHPPerc > HP_PERC_SAFE2ATK) then
   	object = GetMyEnemy_AnyNoKS(MyID) -- get TactBeha and TactSkill too
   	if (object ~= 0) then
         MyState = CHASE_ST
         MyEnemy = object
         TraceAI("IDLE_ST -> CHASE_ST: intercept the detected target")
         return
   	end
   end

   -- [patrol when HP & SP = full] ---------------
   if (CIRCLE_ON_IDLE > 0
   and HomunHP == HomunMaxHP
   and HomunSP == HomunMaxSP
   and (OwnerMotion == MOTION_STAND or OwnerMotion == MOTION_SIT)
   )then
      CircleAroundTarget(GetV(V_OWNER, MyID))
      return
   end

   -- [switch to follow] if owner = on movement --
   local owner_dst = GetDistance2(MyID, GetV(V_OWNER, MyID))
	if (owner_dst > OWNER_CLOSEDISTANCE or owner_dst == -1) then
		MyState = FOLLOW_ST
		MyEnemy = 0
		TraceAI ("IDLE_ST -> FOLLOW_ST: Owner has changed position")
		return
	end

end

--------------------------------------------------
function	OnFOLLOW_ST()
--------------------------------------------------
	-- TraceAI("OnFOLLOW_ST")

   object = GetV(V_TARGET,GetV(V_OWNER, MyID))
   OwnerMotion = GetV(V_MOTION, GetV(V_OWNER, MyID))
   if (object ~= 0
   and (OwnerMotion == MOTION_ATTACK
   or   OwnerMotion == MOTION_ATTACK2)) then
	   TactBeha = GetTact(V_TACTBEHAV, GetV(V_HOMUNTYPE, object))
      TactSkill= GetTact(V_TACTSKILL, GetV(V_HOMUNTYPE, object))
      if (TactBeha ~= BEHA_avoid) then
         MyState = CHASE_ST
         MyEnemy = object
         TraceAI("FOLLOW_ST -> CHASE_ST: cooperative attack")
         return
      end
   end

   local x, y = GetV(V_POSITION, MyID)
   if (math.abs(vHomunX - x) == 0)
   and(math.abs(vHomunY - y) == 0) then         -- if there is no movement
      if (BlockFound == 0) then
         BlockTime = GetTick() + 500            -- 500 ms time-out
         BlockFound = 1
      elseif (BlockFound == 1 and GetTick () > BlockTime) then
         BlockFound = 0
         MoveToOwner(MyID)
		   TraceAI("FOLLOW_ST -> FOLLOW_ST: I was blocked, trying again")
      end
   else
      BlockFound = 0
   end
   vHomunX = x
   vHomunY = y

	if (GetDistanceFromOwner(MyID) <= OWNER_CLOSEDISTANCE) then
		MyState = IDLE_ST
		MyEnemy = 0
		TraceAI("FOLLOW_ST -> IDLE_ST: owner reached")
		return
	elseif (GetV(V_MOTION,MyID) == MOTION_STAND) then
		MoveToOwner(MyID)
		TraceAI("FOLLOW_ST -> FOLLOW_ST: continue to follow owner")
		return
	end

end

--------------------------------------------------
function	OnCHASE_ST()
--------------------------------------------------
	-- TraceAI ("OnCHASE_ST")

   local OwnerID = GetV(V_OWNER, MyID)
	local mob_target = GetV(V_TARGET, MyEnemy)
   local actors = GetActors()

   for i,v in ipairs(actors) do    -- fill the player list
      if (v ~= OwnerID) and (v ~= MyID) then
         if 0 == IsMonster(v) then -- is a player
            if isNotFriend(v) then
               if GetV(V_TARGET, v) == MyEnemy then
                  MyState = IDLE_ST
            	  	MyEnemy = 0
            		TraceAI("CHASE_ST -> IDLE_ST: Another player is aiming at the monster that I was chasing -> Chasing aborted (anti-KS)")
                  return
               end
            end
         end
      end
   end

	if (mob_target > 0) and (mob_target ~=MyID) and (mob_target ~=OwnerID) then
      if isNotFriend(mob_target) then
         MyState = IDLE_ST
         MyEnemy = 0
         TraceAI(string.format("CHASE_ST -> IDLE_ST: MyEnemy was already in battle -> Attack aborted (anti-KS) vs %d (me=%d, owner=%d)", mob_target, MyID, OwnerID))
         return
      end
	end

	local owner_dst = GetDistance2(MyEnemy, OwnerID)
   if (owner_dst >= TOO_FAR_TARGET)then         -- Abort interception if target is too far from the owner
		MyState = FOLLOW_ST
		MyEnemy = 0
		TraceAI("CHASE_ST -> FOLLOW_ST: MyEnemy was too far")
		return
	end

   if (true == IsOutOfSight(MyID, MyEnemy)) then
      MyState = IDLE_ST
		MyEnemy = 0
		MyDestX, MyDestY = 0,0
		TraceAI("CHASE_ST -> IDLE_ST: MyEnemy out of sight")
		return
	end

	if (true == IsInAttackSight(MyID, MyEnemy)) then
		MyState = ATTACK_ST
		TraceAI("CHASE_ST -> ATTACK_ST: MyEnemy in attack range")
		return
	end

	local Ex, Ey = GetV(V_POSITION, MyEnemy)
	local Hx, Hy = GetV(V_POSITION, MyID)

	if (MyDestX ~= Ex or MyDestY ~= Ey) then
		MyDestX, MyDestY = GetV(V_POSITION,MyEnemy)
		Move(MyID, MyDestX, MyDestY)
      TraceAI(string.format("CHASE_ST -> CHASE_ST: target position changed to %d,%d", MyDestX, MyDestY))
		return
	end

   local HomunHP = GetV(V_HP, MyID)
   local HPDiff = HomunHP - vHomunHP
   vHomunHP = HomunHP

   if (HPDiff < 0) then
		MyDestX = Ex
      MyDestY = Ey
		Move(MyID, MyDestX, MyDestY)
      TraceAI("CHASE_ST -> CHASE_ST: hit while chasing, moving againg")
		return
	end

   if (math.abs(vHomunX - Hx) == 0)
   and(math.abs(vHomunY - Hy) == 0) then         -- if there is no movement
      if (BlockFound == 0) then
         BlockTime = GetTick() + 500            -- 1" time-out
         BlockFound = 1
      elseif (BlockFound == 1 and GetTick () > BlockTime) then
         BlockFound = 0
         BlockedPathToEnemyID = MyEnemy
         MyState = FOLLOW_ST
   	  	MyEnemy = 0
	     	MyDestX, MyDestY = GetV(V_POSITION, OwnerID)
		   TraceAI("CHASE_ST -> FOLLOW_ST: Path to target is blocked, searching for another target")
		   return
      end
   else
      BlockFound = 0
   end
   vHomunX = Hx
   vHomunY = Hy

   -- TraceAI(string.format("Chasing... Me=%d,%d Target=%d,%d", Hx, Hy, MyDestX, MyDestY))
end

--------------------------------------------------
function	OnEVADE_ST()
--------------------------------------------------
	-- TraceAI ("OnEVADE_ST")

   -------- evasion-end conditions ---------------
   local EnemyTarget = GetV(V_TARGET, MyEnemy)
   local OwnerID = GetV(V_OWNER, MyID)
   local HomunHP = GetV(V_HP, MyID)
	local HomunMaxHP = GetV(V_MAXHP, MyID)
	local HomunHPPerc = (HomunHP / HomunMaxHP) * 100

	-- 1: if the enemy changes target
   if (EnemyTarget ~= MyID) then
		MyState = IDLE_ST
		MyEnemy = 0
		TraceAI("EVADE_ST -> IDLE: enemy no longer target me")
		return
	end

   -- 2: if the enemy is out of sight
	if (true == IsOutOfSight(MyID, MyEnemy)) then
		MyState = IDLE_ST
		MyEnemy = 0
		TraceAI("EVADE_ST -> IDLE: enemy out of sight")
		return
	end

   -- 3: if the enemy is dead
	if (MOTION_DEAD == GetV(V_MOTION, MyEnemy)) then
		MyState = IDLE_ST
		MyEnemy = 0
		TraceAI("EVADE_ST -> IDLE: enemy dead")
		return
	end

	-- 4: if HP are OK now
   if (HomunHPPerc > HP_PERC_SAFE2ATK) then
      if (TactBeha ~= BEHA_avoid) and (TactBeha ~= BEHA_coward) then
   		MyState = ATTACK_ST
	    	TraceAI("EVADE_ST -> ATTACK: HPs are ok now")
	     	return
	   end
   end

   -- 5: if owner is too far
   local owner_dst = GetDistance2(MyID, OwnerID)
   if (owner_dst >= TOO_FAR_TARGET) then
		MyState = FOLLOW_ST
		MyEnemy = 0
		TraceAI("EVADE_ST -> FOLLOW_ST: I was too far")
		return
	end

   ------- Evading skills ------------------------
	local HomunSP = GetV(V_SP, MyID)
   local type = GetV(V_HOMUNTYPE, MyID)
   if(type == LIF  or type == LIF_H
   or type == LIF2 or type == LIF_H2) then
      DoSkill_UrgentEscape()                    -- Lif: Urgent Escape
   elseif(type == VANILMIRTH  or type == VANILMIRTH_H
   or     type == VANILMIRTH2 or type == VANILMIRTH_H2) then
      DoSkill_CBlessing()                       -- Vani: Chaotic Blessing
   end
   if (HomunHPPerc <= AAP_HP_Perc) then
      if (AAP_Mode == 1 or AAP_Mode == 3) then
         DoSkill_AidPotion(MyID)
      end
   end

   ------- Evading manouvres ---------------------
   local owner_dst = GetDistance2(MyID, OwnerID)
   if (owner_dst >= TOO_FAR_TARGET) then        -- avoid to escape too far
		MyState = FOLLOW_ST
		MyEnemy = 0
		TraceAI("EVADE_ST -> FOLLOW_ST: I was too far")
		return
	end

   CircleAroundTarget(OwnerID)
   CheckForAutoAtk()
   return
end

--------------------------------------------------
function	OnATTACK_ST()
--------------------------------------------------
	-- TraceAI("OnATTACK_ST")

   local OwnerID = GetV(V_OWNER, MyID)
   local type = GetV(V_HOMUNTYPE, MyID)
   local OwnerHP = GetV(V_HP, OwnerID)
	local OwnerMaxHP = GetV(V_MAXHP, OwnerID)
	local OwnerHPPerc = (OwnerHP / OwnerMaxHP) * 100

   -- Heal the Owner, when him/her HPs are low ---
   -- TraceAI(string.format("Owner HPPerc = %d",OwnerHPPerc))
   if (OwnerHPPerc <= AAP_HP_Perc) then
      --   DoSkill_AidPotion(OwnerID)           -- AAP on owner doesn't seem to work: commented out
      if(type == LIF  or type == LIF_H
      or type == LIF2 or type == LIF_H2) then
         DoSkill_HealingTouch(OwnerID)          -- Lif: Healing Touch
      end
   end

   -- Anti-KS: check for the monster target ------
	local mob_target = GetV(V_TARGET, MyEnemy)
	if (mob_target > 0) and (mob_target ~=MyID) and (mob_target ~=OwnerID) then
   	if isNotFriend(mob_target) then
         MyState = IDLE_ST
         MyEnemy = 0
         TraceAI(string.format("ATTACK_ST -> IDLE_ST: MyEnemy was already in battle (anti-KS) vs %d (me=%d, owner=%d)",mob_target, MyID,OwnerID))
         return
      end
	end

   CheckForAutoAtk()

   if (MyEnemy == 0) then
		MyState = IDLE_ST
		TraceAI("ATTACK_ST -> IDLE_ST: no enemy to attack")
		return
	end

	local owner_dst = GetDistance2(MyID, OwnerID)
   if (owner_dst >= TOO_FAR_TARGET) then        -- Abort interception if I'm too far from the owner
		MyState = FOLLOW_ST
		MyEnemy = 0
		TraceAI("ATTACK_ST -> FOLLOW_ST: I was too far")
		return
	end

   -- Get current status information -------------
   local HomunHP     = GetV(V_HP,MyID)
	local HomunMaxHP  = GetV(V_MAXHP,MyID)
	local HomunSP     = GetV(V_SP,MyID)
	local HomunMaxSP  = GetV(V_MAXSP,MyID)
   local EnemyTarget = GetV(V_TARGET, MyEnemy)
   local HomunHPPerc = (HomunHP / HomunMaxHP) * 100

   -- Heal the homunculus with AAP, if he/she is in danger
   if (HomunHPPerc <= AAP_HP_Perc) then
      if (AAP_Mode == 2 or AAP_Mode == 3) then
         DoSkill_AidPotion(MyID)
      end
   end

--[[
   [...]
   -- Make sure no other monster are attacking ---
   if (EnemyTarget ~= MyID) then                -- if current enemy is not aiming at the homunculus
      object = GetMyEnemy_AttackingMe(MyID)     -- check if another monster is " " " in the meantime:
      if (object ~= 0) then                     -- if a threat has been detected
         MyEnemy = object                       -- change target
         AtkSkillDoneCount = 0
         AtkStartTime = 0                       -- a new attack starts (usually this happen in ON_IDLE_ST)
      end
   end
--]]

   -- Survival Instinct --------------------------
   if
      ((EnemyTarget == MyID) and (HomunHPPerc < HP_PERC_DANGER))
   or (TactBeha == BEHA_avoid)
   then
      MyState = EVADE_ST
      Evade_ComeBackTime = EVADE_COMEBACK
      TraceAI("ATTACK_ST -> EVADE_ST: HP < HP_PERC_DANGER or beha=avoid")
      return
   end

   -- Stop if the enemy is out of sight ----------
   if (true == IsOutOfSight(MyID, MyEnemy)) then
      MyState = IDLE_ST
      MyEnemy = 0
      TraceAI("ATTACK_ST -> IDLE_ST: enemy out of sight or dead")
      return
   end

   -- Stop if the enemy is dead ------------------
   if (MOTION_DEAD == GetV(V_MOTION, MyEnemy)) then
      MyState = IDLE_ST
      MyEnemy = 0
      TraceAI("ATTACK_ST -> IDLE_ST: enemy dead")
      return
   end

   -- Intercept the enemy ------------------------
   if (false == IsInAttackSight(MyID,MyEnemy)) then
      MyState = CHASE_ST
      MyDestX, MyDestY = GetV(V_POSITION, MyEnemy)
      Move(MyID, MyDestX, MyDestY)
      TraceAI("ATTACK_ST -> CHASE_ST: enemy out of attack range")
      return
   end

   -- Autoskill system ---------------------------
   local EnemyType = GetV(V_HOMUNTYPE, MyEnemy)
   local CurrTime = GetTick()
   if (AtkStartTime == 0) then
      AtkStartTime = CurrTime
   end
   if (MySkill == 0) then
      Attack(MyID, MyEnemy)

      if (EnemyType < 1078 or EnemyType > 1085) then -- don't waste SP on plants and mushrooms

         local CanUseAtkSkills = true -- is it the right time for an aggressive skill?
         local TimeOutElapsed = CurrTime - AtkStartTime > SKILL_TIME_OUT
         if (TactSkill == WITH_no_skill) then
            CanUseAtkSkills = false
         elseif (TactSkill == WITH_one_skill) then
            if( (AtkSkillDoneCount > 1) or  TimeOutElapsed)then
               CanUseAtkSkills = false
            end
         elseif (TactSkill == WITH_two_skills) then
            if( (AtkSkillDoneCount > 2) or  TimeOutElapsed)then
               CanUseAtkSkills = false
            end
         elseif (TactSkill == WITH_max_skills) then
            if (TimeOutElapsed) then
               AtkSkillTimeOK = false
            end
         elseif (TactSkill == WITH_slow_power) then
            if (CurrTime - AtkStartTime < DELAY_SLOW_POWER) then
               CanUseAtkSkills = false
            end
         end

         -- Amistr ----------------------------------
         if(type == AMISTR  or type == AMISTR_H
         or     type == AMISTR2 or type == AMISTR_H2) then
         --------------------------------------------

            -- Bulwark
            if (AS_AMI_BULW_Engaged) then
               if (GetTick() > AS_AMI_BULW_TimeOut) then
                  AS_AMI_BULW_Engaged = false
               end
            else
               if (HomunSP >= AUTOSKILL_SP_AMI_BULW) then
                  MySkill = 8006
                  MySkillLevel = 5
                  AS_AMI_BULW_TimeOut = GetTick() + DELAY_AMI_BULW
                  AS_AMI_BULW_Engaged = true
               end
            end

         -- Lif -------------------------------------
         elseif(type == LIF  or type == LIF_H
         or     type == LIF2 or type == LIF_H2) then
         --------------------------------------------

            DoSkill_UrgentEscape()

         -- Filir -----------------------------------
         elseif (type == FILIR  or type == FILIR_H
         or      type == FILIR2 or type == FILIR_H2) then
         --------------------------------------------

            -- Accelerated flight
            if (AS_FIL_ACCL_Engaged) then          -- if AF is active, wait for its time-out before to activate AF again
               if (GetTick() > AS_FIL_ACCL_TimeOut) then
                  AS_FIL_ACCL_Engaged = false
               end
            else                                   -- if AF is OFF, activate it
               if (HomunSP >= AUTOSKILL_SP_FIL_ACCL) then -- if there are enough SP left
                  MySkill = 8011
                  MySkillLevel = 5
                  AS_FIL_ACCL_TimeOut = GetTick() + DELAY_FIL_ACCL
                  AS_FIL_ACCL_Engaged = true
               end
            end

            -- Moonlight <--------- ##### AGGRESSIVE SKILL #####
            if (HomunSP >= AUTOSKILL_SP_FIL_MOON) then
               if (CanUseAtkSkills) then
                  MySkill = 8009
                  MySkillLevel = 5
                  AtkSkillDoneCount = AtkSkillDoneCount + 1
               end
            end

         -- Vanilmirth ------------------------------
         elseif(type == VANILMIRTH  or type == VANILMIRTH_H
         or     type == VANILMIRTH2 or type == VANILMIRTH_H2) then
         --------------------------------------------

            -- Caprice <----------- ##### AGGRESSIVE SKILL #####
            if (HomunSP >= AUTOSKILL_SP_VAN_CAPR) then
               if (CanUseAtkSkills) then
                  MySkill = 8013
                  MySkillLevel = 5
                  AtkSkillDoneCount = AtkSkillDoneCount + 1
               end
            end

         end

      end

   end
   if (MySkill ~= 0) then
      SkillObject(MyID, MySkillLevel, MySkill, MyEnemy)
      MySkill = 0
   end

   return
end

--------------------------------------------------
function	OnOUTMOVE_NEWFRIEND_ST()
--------------------------------------------------
   local CurrTime = GetTick() - OutMoveTime
   if CurrTime > 3000 then
   	MyState = FOLLOW_ST
   	MyEnemy = 0
   else
      local OutMoveX, OutMoveY = GetV(V_POSITION, OutMoveTarget)
      local angle = math.rad(math.mod(CurrTime/4, 360))
      local X = OutMoveX + math.cos(angle) * 3
      local Y = OutMoveY + math.sin(angle) * 3
      Move(MyID, X, Y)
   	MyDestX = X
   	MyDestY = Y
   end
end

--------------------------------------------------
function	OnOUTMOVE_DELFRIEND_ST()
--------------------------------------------------
   local CurrTime = GetTick() - OutMoveTime
   if CurrTime > 3000 then
   	MyState = FOLLOW_ST
   	MyEnemy = 0
   else
      local OutMoveX, OutMoveY = GetV(V_POSITION, OutMoveTarget)
      local angle = math.rad(math.mod(CurrTime/4, 360))
      local X = OutMoveX + math.cos(angle) * 3
      local Y = OutMoveY
      Move(MyID, X, Y)
   	MyDestX = X
   	MyDestY = Y
   end
end

--------------------------------------------------
function	OnMOVE_CMD_ST()
--------------------------------------------------
	TraceAI("OnMOVE_CMD_ST")

	local x, y = GetV (V_POSITION,MyID)
	if (x == MyDestX and y == MyDestY) then				-- DESTINATION_ARRIVED_IN
		MyState = IDLE_ST
		MyEnemy = 0
	end
end

--------------------------------------------------
function OnSTOP_CMD_ST()
--------------------------------------------------
end

--------------------------------------------------
function OnATTACK_OBJECT_CMD_ST()
--------------------------------------------------
end

--------------------------------------------------
function OnATTACK_AREA_CMD_ST()
--------------------------------------------------
	TraceAI ("OnATTACK_AREA_CMD_ST")

	local	object = GetOwnerEnemy(MyID)
	if (object == 0) then
		object = GetMyEnemy_AnyNoKS(MyID)
	end

	if (object ~= 0) then							-- MYOWNER_ATTACKED_IN or ATTACKED_IN
		MyState = CHASE_ST
		MyEnemy = object
		return
	end

	local x , y = GetV (V_POSITION,MyID)
	if (x == MyDestX and y == MyDestY) then			-- DESTARRIVED_IN
		MyState = IDLE_ST
		MyEnemy = 0
	end

end

--------------------------------------------------
function OnPATROL_CMD_ST()
--------------------------------------------------
	TraceAI("OnPATROL_CMD_ST")

	local	object = GetOwnerEnemy(MyID)
	if (object == 0) then
		object = GetMyEnemy_AnyNoKS(MyID)
	end

	if (object ~= 0) then							-- MYOWNER_ATTACKED_IN or ATTACKED_IN
		MyState = CHASE_ST
		MyEnemy = object
		TraceAI("PATROL_CMD_ST -> CHASE_ST : ATTACKED_IN")
		return
	end

	local x , y = GetV (V_POSITION,MyID)
	if (x == MyDestX and y == MyDestY) then			-- DESTARRIVED_IN
		MyDestX = MyPatrolX
		MyDestY = MyPatrolY
		MyPatrolX = x
		MyPatrolY = y
		Move (MyID,MyDestX,MyDestY)
	end

end

--------------------------------------------------
function OnHOLD_CMD_ST ()
--------------------------------------------------
	TraceAI("OnHOLD_CMD_ST")

	if (MyEnemy ~= 0) then
		local d = GetDistance(MyEnemy,MyID)
		if (d ~= -1 and d <= GetV(V_ATTACKRANGE,MyID)) then
				Attack (MyID,MyEnemy)
		else
			MyEnemy = 0
		end
		return
	end

	local	object = GetOwnerEnemy(MyID)
	if (object == 0) then
		object = GetMyEnemy_AnyNoKS(MyID)
		if (object == 0) then
			return
		end
	end

	MyEnemy = object

end

--------------------------------------------------
function OnSKILL_OBJECT_CMD_ST()
--------------------------------------------------
end

--------------------------------------------------
function OnSKILL_AREA_CMD_ST()
--------------------------------------------------
	TraceAI("OnSKILL_AREA_CMD_ST")

	local x , y = GetV (V_POSITION,MyID)
	if (GetDistance(x,y,MyDestX,MyDestY) <= GetV(V_SKILLATTACKRANGE,MyID,MySkill)) then	-- DESTARRIVED_IN
		SkillGround (MyID,MySkillLevel,MySkill,MyDestX,MyDestY)
		MyState = IDLE_ST
		MySkill = 0
	end

end

--------------------------------------------------
function OnFOLLOW_CMD_ST ()
--------------------------------------------------
	TraceAI("OnFOLLOW_CMD_ST")

	local ownerX, ownerY, myX, myY
	ownerX, ownerY = GetV (V_POSITION,GetV(V_OWNER,MyID))
	myX, myY = GetV (V_POSITION,MyID)

	local d = GetDistance (ownerX,ownerY,myX,myY)

	if ( d <= 3) then
		return
	end

	local motion = GetV (V_MOTION,MyID)
	if (motion == MOTION_MOVE) then
		d = GetDistance (ownerX, ownerY, MyDestX, MyDestY)
		if ( d > 3) then
			MoveToOwner (MyID)
			MyDestX = ownerX
			MyDestY = ownerY
			return
		end
	else
		MoveToOwner (MyID)
		MyDestX = ownerX
		MyDestY = ownerY
		return
	end

end

--------------------------------------------------
function	CheckForAutoAtk()
--------------------------------------------------
   -- Alchemist auto attacks ---------------------
   -- Cart Revolution will be engaged if there are 2 or more enemy close to the owner
   local NOE = CountEnemiesCloseToOwner() -- Near Owner Enemies
	-- TraceAI(string.format("Enemies close to owner: %d", NOE))
   if NOE > 0 then
      if (ACR_MinEnemies > 0 and NOE >= ACR_MinEnemies) then
         DoSkill_AutoAttack(153)
      else
         if (AST_SkillID > 0) then
            DoSkill_AutoAttack(AST_SkillID)
         end
      end
   end
end

--------------------------------------------------
function	GetOwnerEnemy(myid)
--------------------------------------------------
	local result = 0
	local owner  = GetV(V_OWNER, myid)
	local actors = GetActors()
	local enemys = {}
	local index = 1
	local target
	for i,v in ipairs(actors) do
		if (v ~= owner and v ~= myid) then
			target = GetV(V_TARGET, v)
			if (target == owner) then
				if (IsMonster(v) == 1) then
					enemys[index] = v
					index = index + 1
				else
					local motion = GetV(V_MOTION, i)
					if (motion == MOTION_ATTACK or motion == MOTION_ATTACK2) then
						enemys[index] = v
						index = index + 1
					end
				end
			end
		end
	end

	local min_dis = 100
	local dis
	for i,v in ipairs(enemys) do
		dis = GetDistance2 (myid,v)
		if (dis < min_dis) then
			result = v
			min_dis = dis
		end
	end

	return result
end

--------------------------------------------------
function	GetMyEnemy_DefaultAI(myid)
--------------------------------------------------
	local result = 0

	local type = GetV(V_HOMUNTYPE, myid)
	if (type == LIF    or type == LIF_H
   or type == AMISTR  or type == AMISTR_H
   or type == LIF2    or type == LIF_H2
   or type == AMISTR2 or type == AMISTR_H2) then
		result = GetMyEnemy_AttackingMe(myid)
	elseif (type == FILIR  or type == FILIR_H
   or type == VANILMIRTH  or type == VANILMIRTH_H
   or type == FILIR2      or type == FILIR_H2
   or type == VANILMIRTH2 or type == VANILMIRTH_H2) then
		result = GetMyEnemy_AnynoKS(myid)
	end
	return result
end

--------------------------------------------------
function CountEnemiesCloseToOwner()
--------------------------------------------------
	local OwnerID = GetV(V_OWNER, MyID)
	local ecount = 0
	local actors = GetActors()
	for i,v in ipairs(actors) do
		if (v ~= OwnerID and v ~= MyID) then
   		if (GetDistance2(OwnerID, v) <= 1) then
            ecount = ecount + 1
   		end
		end
	end
	return ecount
end

--------------------------------------------------
function	GetMyEnemy_AttackingMe(myid)
--------------------------------------------------
   -- TraceAI(string.format("---- PASSIVE SCAN STARTED for ID (%d)", myid))
	local result = 0
	local actors = GetActors()
	local aggressors = {}
	local aggr_index = 1
	local target
	local OwnerID = GetV(V_OWNER, MyID)

   -- Fill a list of enemies who have me as target
	for i,v in ipairs(actors) do
      if GetV(V_MOTION, v) ~= MOTION_DEAD then
         if (v ~= OwnerID) and (v ~= MyID) then
            target = GetV(V_TARGET, v)
            if (target == MyID) then
               aggressors[aggr_index] = v
               aggr_index = aggr_index + 1
            end
         end
		end
	end

   -- Search for the closest aggressor -----------
	local nearest_dst = 100
	local owner_dst = 0
	local dst
	local MobType
	local tmpBEHA
	local lastweak = 0
	for i,v in ipairs(aggressors) do
		dst = GetDistance2(MyID, v)
		owner_dst = GetDistance2(OwnerID, v)
		if (dst < nearest_dst and owner_dst < TOO_FAR_TARGET) then
			nearest_dst = dst
			MobType = GetV(V_HOMUNTYPE, v)
			tmpBEHA = GetTact(V_TACTBEHAV, MobType)
			if tmpBEHA == BEHA_attack_weak then -- ignore futile attacks from weak monsters
			   lastweak = v
			else
			   result = v
			end
         if (tmpBEHA == BEHA_avoid) or (tmpBEHA == BEHA_coward) then
            break -- this monster is too dangerous: evade from him/her at once!
         end
		end
	end

   if (result == 0) and (lastweak ~=0) and (MyState == IDLE_ST) then
      -- TraceAI("Only weak targets are hitting me")
      local HomunHP = GetV(V_HP, MyID)
      local HomunMaxHP = GetV(V_MAXHP, MyID)
   	local HomunHPPerc = (HomunHP / HomunMaxHP) * 100
      if (HomunHPPerc <= HP_PERC_SAFE2ATK) then
         -- No real threats were found, but some weak (= BEHA_attack_weak) enemies are hitting me: usually the
         -- aggressive scan (GetMyEnemy_AnyNoKS function) takes care of them (in the correct priority order), but
         -- now the homunculus HPs are under the safe level, so he/she is in defensive-only mode: no agressive
         -- scan will be started at this time! So it's ok to attack now this low priority monster
         result = lastweak
      end
   end
   if result ~=0 then
      -- TraceAI("Monster attacking me found:")
      TactBeha = tmpBEHA
      TactSkill= GetTact(V_TACTSKILL, MobType)
   end

   -- TraceAI("---- END OF PASSIVE SCAN")
	return result
end

--------------------------------------------------
function GetClosest(EnemiesOnSight)
--------------------------------------------------
   local result = 0
   local OwnerID = GetV(V_OWNER, MyID)
   local nearest_dst = 100
   local owner_dst = 0
   local dst = 0
   for i,v in ipairs(EnemiesOnSight) do
      dst = GetDistance2(MyID, v)
      -- TraceAI(string.format("distance from EnemyID %d is %d", v, dst))
      owner_dst = GetDistance2(OwnerID, v)
      if (dst < nearest_dst and owner_dst < TOO_FAR_TARGET) then
         result = v
         nearest_dst = dst
         -- TraceAI(string.format("he/she is now the closest (%d)", nearest_dst))
      end
   end
	return result
end

--------------------------------------------------
function	GetMyEnemy_AnyNoKS(myid)
--------------------------------------------------
   -- TraceAI(string.format("---- AGGRESSIVE SCAN STARTED for ID (%d)", myid))
   local result = 0
   local OwnerID = GetV(V_OWNER, MyID)
   local actors = GetActors ()
   local players = {}
   local player_idx = 1
   local enemy_1st = {} -- primary targets
   local index_1st = 1
   local enemy_std = {} -- normal targets
   local index_std = 1
   local enemy_lss = {} -- last targets
   local index_lss = 1
   local mob_type
   local mob_target
   local mobBeha
   local ListedType
	local EnemyMaxHP
	local HomunMaxHP = GetV(V_MAXHP, MyID)

   for i,v in ipairs(actors) do -- fill the (not "friend") player list
      if (v ~= OwnerID and v ~= MyID) then
         if (0 == IsMonster(v)) then
            if isNotFriend(v) then
               players[player_idx] = v
               player_idx = player_idx + 1
            end
         end
      end
   end

	for i,v in ipairs(actors) do
      if (v ~= OwnerID) and (v ~= MyID) then      -- owner and homunculus aren't valid target, of course
         if (1 == IsMonster(v)) and (BlockedPathToEnemyID ~= v) then -- if target is a (reachable) monster
         if GetV(V_MOTION, v) ~= MOTION_DEAD then -- and hopefully still alive...

            local isOKToAttack = false
            mob_type = GetV(V_HOMUNTYPE, v)       -- get tact data
            mobBeha = GetTact(V_TACTBEHAV, mob_type)
            if( (mobBeha == BEHA_attack_1st)
            or  (mobBeha == BEHA_attack)
            or  (mobBeha == BEHA_attack_last)
            or  (mobBeha == BEHA_attack_weak)
            ) then                                -- OK, this type of monster is to attack
               mob_target = GetV(V_TARGET, v)     -- check for the monster's target
               if (mob_target == 0) then          -- if the monster is not in battle
                  isOKToAttack = true             -- he/she can be attacked, unless...
                  for i2,v2 in ipairs(players) do
                     if (v == GetV(V_TARGET, v2)) then
                        isOKToAttack = false      -- ...another player is already aiming at him/her (anti-KS)
                        break
                     end
                  end
               else                               -- else, if the monster is already in battle
                  if (mob_target == MyID)
                  or (mob_target == OwnerID)
                  or (not isNotFriend(mob_target)) then
                     isOKToAttack = true          -- but with me/my owner/a friend, it's OK (anti-KS)
                  end
               end
            end

            if isOKToAttack then
               if (mobBeha == BEHA_attack_1st) then
                  enemy_1st[index_1st] = v
                  index_1st = index_1st + 1
                  -- TraceAI(string.format(" added to (1st) as enemyID %d", v))
               elseif (mobBeha == BEHA_attack) then
                  enemy_std[index_std] = v
                  index_std = index_std + 1
                  -- TraceAI(string.format(" added to (std) as enemyID %d", v))
               else -- BEHA_attack_last, BEHA_attack_weak
                  enemy_lss[index_lss] = v
                  index_lss = index_lss + 1
                  -- TraceAI(string.format(" added to (lss) as enemyID %d", v))
               end
            end

         end
         end
		end
	end

   -- Searches for nearest target ---------------
   result = GetClosest(enemy_1st)
   if (result == 0) then
      -- TraceAI("No primary target in range")
      result = GetClosest(enemy_std)
   end
   if (result == 0) then
      -- TraceAI("No standard target in range")
      result = GetClosest(enemy_lss)
   end
   if (result == 0) then
      -- TraceAI("No target in range")
   else
      TraceAI(string.format("Target found (ID %d):", result))
      mob_type = GetV(V_HOMUNTYPE, result)
   	TactBeha = GetTact(V_TACTBEHAV, mob_type)
      TactSkill= GetTact(V_TACTSKILL, mob_type)
   end

   BlockedPathToEnemyID = 0
   -- TraceAI("---- END OF AGGRESSIVE SCAN")
	return result
end

--------------------------------------------------
function AI(myid)
--------------------------------------------------
	MyID = myid
	local msg	= GetMsg (myid)			-- command
	local rmsg	= GetResMsg (myid)		-- reserved command

   if not FriendList_Loaded then
      FriendList_Load()
      FriendList_Loaded = true
   end

	if msg[1] == NONE_CMD then
		if rmsg[1] ~= NONE_CMD then
			if List.size(ResCmdList) < 10 then
				List.pushright (ResCmdList,rmsg)
			end
		end
	else
		List.clear (ResCmdList)
		ProcessCommand (msg)
	end

 	if     (MyState == IDLE_ST) then
		OnIDLE_ST ()
	elseif (MyState == CHASE_ST) then
		OnCHASE_ST ()
	elseif (MyState == EVADE_ST) then
		OnEVADE_ST ()
	elseif (MyState == ATTACK_ST) then
		OnATTACK_ST ()
	elseif (MyState == FOLLOW_ST) then
		OnFOLLOW_ST ()
   -----------------------------------
   elseif (MyState == OUTMOVE_NEWFRIEND_ST) then
		OnOUTMOVE_NEWFRIEND_ST()
   elseif (MyState == OUTMOVE_DELFRIEND_ST) then
		OnOUTMOVE_DELFRIEND_ST()
   -----------------------------------
	elseif (MyState == MOVE_CMD_ST) then
		OnMOVE_CMD_ST ()
	elseif (MyState == STOP_CMD_ST) then
		OnSTOP_CMD_ST ()
	elseif (MyState == ATTACK_OBJECT_CMD_ST) then
		OnATTACK_OBJECT_CMD_ST ()
	elseif (MyState == ATTACK_AREA_CMD_ST) then
		OnATTACK_AREA_CMD_ST ()
	elseif (MyState == PATROL_CMD_ST) then
		OnPATROL_CMD_ST ()
	elseif (MyState == HOLD_CMD_ST) then
		OnHOLD_CMD_ST ()
	elseif (MyState == SKILL_OBJECT_CMD_ST) then
		OnSKILL_OBJECT_CMD_ST ()
	elseif (MyState == SKILL_AREA_CMD_ST) then
		OnSKILL_AREA_CMD_ST ()
	elseif (MyState == FOLLOW_CMD_ST) then
		OnFOLLOW_CMD_ST ()
	end

end
